import json
import os
import shutil
import subprocess
import sys


def finish(**kwargs):
    json.dump(kwargs, sys.stdout)
    sys.stdout.flush()


def run_command(cmd, env, stdin=None):
    if stdin is not None:
        stdin = stdin.encode("utf8")
        STDIN = subprocess.PIPE
    else:
        STDIN = None

    proc = subprocess.Popen(
        cmd,
        env=env,
        cwd=os.path.expanduser("~"),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=STDIN,
    )

    stdout, stderr = proc.communicate(stdin)

    finish(
        ok=True,
        returncode=proc.returncode,
        stdout=stdout.decode("utf8", "replace"),
        stderr=stderr.decode("utf8", "replace"),
    )


def start(cmd, env, stdin=None, staging_dir=None, files=None):
    if staging_dir:
        try:
            os.makedirs(staging_dir, mode=0o700, exist_ok=False)
            for name, value in files.items():
                with open(os.path.join(staging_dir, name), "w") as f:
                    f.write(value)
        except Exception as exc:
            finish(
                ok=False,
                error=f"Error setting up staging directory {staging_dir}: {exc}",
            )
            return
    run_command(cmd, env, stdin=stdin)


def stop(cmd, env, staging_dir=None):
    if staging_dir:
        if not os.path.exists(staging_dir):
            return
        try:
            shutil.rmtree(staging_dir)
        except Exception as exc:
            finish(
                ok=False,
                error=f"Error removing staging directory {staging_dir}: {exc}",
            )
            return
    run_command(cmd, env)


def main():
    try:
        kwargs = json.load(sys.stdin)
    except ValueError as exc:
        finish(ok=False, error=str(exc))
        return

    action = kwargs.pop("action", None)
    if action == "start":
        start(**kwargs)
    elif action == "stop":
        stop(**kwargs)
    else:
        finish(ok=False, error="Valid actions are 'start' and 'stop'")


if __name__ == "__main__":
    main()



import asyncio

import aiohttp
from traitlets import (
    Dict,
    Float,
    Instance,
    Integer,
    Unicode,
    Union,
    default,
    observe,
    validate,
)
from traitlets.config import Configurable, LoggingConfigurable

from .. import models
from ..options import Options
from ..traitlets import Callable, Command, MemoryLimit, Type
from ..utils import awaitable, format_bytes

__all__ = ("Backend", "ClusterConfig")


class PublicException(Exception):
    """An exception that can be reported to the user"""

    pass


class Backend(LoggingConfigurable):
    """Base class for defining dask-gateway backends.

    Subclasses should implement the following methods:

    - ``setup``
    - ``cleanup``
    - ``start_cluster``
    - ``stop_cluster``
    - ``on_cluster_heartbeat``
    """

    cluster_options = Union(
        [Callable(), Instance(Options, args=())],
        help="""
        User options for configuring an individual cluster.

        Allows users to specify configuration overrides when creating a new
        cluster. See the documentation for more information:

        :doc:`cluster-options`.
        """,
        config=True,
    )

    cluster_config_class = Type(
        "dask_gateway_server.backends.base.ClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    scheduler_api_retries = Integer(
        3,
        min=0,
        help="""
        The number of attempts to make when contacting a scheduler api endpoint.

        If failures occur after the max number of retries, the dask cluster will
        be marked as failed and will be cleaned up.
        """,
    )

    api_url = Unicode(
        help="""
        The address that internal components (e.g. dask clusters)
        will use when contacting the gateway.
        """,
        config=True,
    )

    # Forwarded from the main application
    gateway_address = Unicode()

    @default("gateway_address")
    def _gateway_address_default(self):
        return self.parent.address

    async def get_cluster_options(self, user):
        if callable(self.cluster_options):
            return await awaitable(self.cluster_options(user))
        return self.cluster_options

    async def process_cluster_options(self, user, request):
        try:
            cluster_options = await self.get_cluster_options(user)
            requested_options = cluster_options.parse_options(request)
            overrides = cluster_options.get_configuration(requested_options, user)
            config = self.cluster_config_class(parent=self, **overrides)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            raise PublicException(str(exc))
        return requested_options, config

    async def forward_message_to_scheduler(self, cluster, msg):
        if cluster.status != models.ClusterStatus.RUNNING:
            raise PublicException(f"cluster {cluster.name} is not running")
        attempt = 1
        t = 0.1
        while True:
            try:
                await self.session.post(
                    cluster.api_address + "/api/comm",
                    json=msg,
                    headers={"Authorization": "token %s" % cluster.token},
                    raise_for_status=True,
                )
                return
            except Exception:
                if attempt < self.scheduler_api_retries:
                    self.log.warning(
                        f"Failed to message cluster {cluster.name} on attempt "
                        f"{attempt}, retrying in {t} s",
                        exc_info=True,
                    )
                    await asyncio.sleep(t)
                    attempt += 1
                    t = min(t * 2, 5)
                else:
                    break
        self.log.warning(
            f"Failed to message cluster {cluster.name} on attempt "
            f"{attempt}, marking cluster as failed"
        )
        await self.stop_cluster(cluster.name, failed=True)
        raise PublicException(f"cluster {cluster.name} is not running")

    async def setup(self, app):
        """Called when the server is starting up.

        Do any initialization here.

        Parameters
        ----------
        app : aiohttp.web.Application
            The aiohttp application. Can be used to add additional routes if
            needed.
        """
        self.session = aiohttp.ClientSession()

    async def cleanup(self):
        """Called when the server is shutting down.

        Do any cleanup tasks in this method"""
        await self.session.close()

    async def list_clusters(self, username=None, statuses=None):
        """List known clusters.

        Parameters
        ----------
        username : str, optional
            A user name to filter on. If not provided, defaults to
            all users.
        statuses : list, optional
            A list of statuses to filter on. If not provided, defaults to all
            running and pending clusters.

        Returns
        -------
        clusters : List[Cluster]
        """
        raise NotImplementedError

    async def get_cluster(self, cluster_name, wait=False):
        """Get information about a cluster.

        Parameters
        ----------
        cluster_name : str
            The cluster name.
        wait : bool, optional
            If True, wait until the cluster has started or failed before
            returning. If waiting is not possible (or waiting for a long period
            of time would be expensive) it is valid to return early with a
            Cluster object in a state prior to RUNNING (the client will retry
            in this case). Default is False.

        Returns
        -------
        cluster : Cluster
        """
        raise NotImplementedError

    async def start_cluster(self, user, cluster_options):
        """Submit a new cluster.

        Parameters
        ----------
        user : User
            The user making the request.
        cluster_options : dict
            Any additional options provided by the user.

        Returns
        -------
        cluster_name : str
        """
        raise NotImplementedError

    async def stop_cluster(self, cluster_name, failed=False):
        """Stop a cluster.

        No-op if the cluster is already stopped.

        Parameters
        ----------
        cluster_name : str
            The cluster name.
        failed : bool, optional
            If True, the cluster should be marked as FAILED after stopping. If
            False (default) it should be marked as STOPPED.
        """
        raise NotImplementedError

    async def on_cluster_heartbeat(self, cluster_name, msg):
        """Handle a cluster heartbeat.

        Parameters
        ----------
        cluster_name : str
            The cluster name.
        msg : dict
            The heartbeat message.
        """
        raise NotImplementedError


class ClusterConfig(Configurable):
    """Base class for holding individual Dask cluster configurations"""

    scheduler_cmd = Command(
        "dask-scheduler", help="Shell command to start a dask scheduler.", config=True
    )

    worker_cmd = Command(
        "dask-worker", help="Shell command to start a dask worker.", config=True
    )

    environment = Dict(
        help="""
        Environment variables to set for both the worker and scheduler processes.
        """,
        config=True,
    )

    worker_memory = MemoryLimit(
        "2 G",
        help="""
        Number of bytes available for a dask worker. Allows the following
        suffixes:

        - K -> Kibibytes
        - M -> Mebibytes
        - G -> Gibibytes
        - T -> Tebibytes
        """,
        config=True,
    )

    worker_cores = Integer(
        1,
        min=1,
        help="""
        Number of cpu-cores available for a dask worker.
        """,
        config=True,
    )

    # Number of threads per worker. Defaults to the number of cores
    worker_threads = Integer(
        help="""
        Number of threads available for a dask worker.

        Defaults to ``worker_cores``.
        """,
        min=1,
        config=True,
        allow_none=True,
    )

    @default("worker_threads")
    def _default_worker_threads(self):
        return max(int(self.worker_cores), 1)

    @validate("worker_threads")
    def _validate_worker_threads(self, proposal):
        if not proposal.value:
            return self._default_worker_threads()
        return proposal.value

    scheduler_memory = MemoryLimit(
        "2 G",
        help="""
        Number of bytes available for a dask scheduler. Allows the following
        suffixes:

        - K -> Kibibytes
        - M -> Mebibytes
        - G -> Gibibytes
        - T -> Tebibytes
        """,
        config=True,
    )

    scheduler_cores = Integer(
        1,
        min=1,
        help="""
        Number of cpu-cores available for a dask scheduler.
        """,
        config=True,
    )

    adaptive_period = Float(
        3,
        min=0,
        help="""
        Time (in seconds) between adaptive scaling checks.

        A smaller period will decrease scale up/down latency when responding to
        cluster load changes, but may also result in higher load on the gateway
        server.
        """,
        config=True,
    )

    idle_timeout = Float(
        0,
        min=0,
        help="""
        Time (in seconds) before an idle cluster is automatically shutdown.

        Set to 0 (default) for no idle timeout.
        """,
        config=True,
    )

    cluster_max_memory = MemoryLimit(
        None,
        help="""
        The maximum amount of memory (in bytes) available to this cluster.
        Allows the following suffixes:

        - K -> Kibibytes
        - M -> Mebibytes
        - G -> Gibibytes
        - T -> Tebibytes

        Set to ``None`` for no memory limit (default).
        """,
        min=0,
        allow_none=True,
        config=True,
    )

    cluster_max_cores = Float(
        None,
        help="""
        The maximum number of cores available to this cluster.

        Set to ``None`` for no cores limit (default).
        """,
        min=0.0,
        allow_none=True,
        config=True,
    )

    cluster_max_workers = Integer(
        help="""
        The maximum number of workers available to this cluster.

        Note that this will be combined with ``cluster_max_cores`` and
        ``cluster_max_memory`` at runtime to determine the actual maximum
        number of workers available to this cluster.
        """,
        allow_none=True,
        min=0,
        config=True,
    )

    def _check_scheduler_memory(self, scheduler_memory, cluster_max_memory):
        if cluster_max_memory is not None and scheduler_memory > cluster_max_memory:
            memory = format_bytes(scheduler_memory)
            limit = format_bytes(cluster_max_memory)
            raise ValueError(
                f"Scheduler memory request of {memory} exceeds cluster memory "
                f"limit of {limit}"
            )

    def _check_scheduler_cores(self, scheduler_cores, cluster_max_cores):
        if cluster_max_cores is not None and scheduler_cores > cluster_max_cores:
            raise ValueError(
                f"Scheduler cores request of {scheduler_cores} exceeds cluster "
                f"cores limit of {cluster_max_cores}"
            )

    def _worker_limit_from_resources(self):
        inf = max_workers = float("inf")
        if self.cluster_max_memory is not None:
            max_workers = min(
                (self.cluster_max_memory - self.scheduler_memory) // self.worker_memory,
                max_workers,
            )
        if self.cluster_max_cores is not None:
            max_workers = min(
                (self.cluster_max_cores - self.scheduler_cores) // self.worker_cores,
                max_workers,
            )

        if max_workers == inf:
            return None
        return max(0, int(max_workers))

    @validate("scheduler_memory")
    def _validate_scheduler_memory(self, proposal):
        self._check_scheduler_memory(proposal.value, self.cluster_max_memory)
        return proposal.value

    @validate("scheduler_cores")
    def _validate_scheduler_cores(self, proposal):
        self._check_scheduler_cores(proposal.value, self.cluster_max_cores)
        return proposal.value

    @validate("cluster_max_memory")
    def _validate_cluster_max_memory(self, proposal):
        self._check_scheduler_memory(self.scheduler_memory, proposal.value)
        return proposal.value

    @validate("cluster_max_cores")
    def _validate_cluster_max_cores(self, proposal):
        self._check_scheduler_cores(self.scheduler_cores, proposal.value)
        return proposal.value

    @validate("cluster_max_workers")
    def _validate_cluster_max_workers(self, proposal):
        lim = self._worker_limit_from_resources()
        if lim is None:
            return proposal.value
        if proposal.value is None:
            return lim
        return min(proposal.value, lim)

    @observe("cluster_max_workers")
    def _observe_cluster_max_workers(self, change):
        # This shouldn't be needed, but traitlet validators don't run
        # if a value is `None` and `allow_none` is true, so we need to
        # add an observer to handle the event of an *explicit* `None`
        # set for `cluster_max_workers`
        if change.new is None:
            lim = self._worker_limit_from_resources()
            if lim is not None:
                self.cluster_max_workers = lim

    @default("cluster_max_workers")
    def _default_cluster_max_workers(self):
        return self._worker_limit_from_resources()

    def to_dict(self):
        return {
            k: getattr(self, k)
            for k in self.trait_names()
            if k not in {"parent", "config"}
        }



"""
dask-gateway-server has Golang source code in ./dask-gateway-proxy that we want
to compile as part of installing this Package and building a source distribution
or wheels. There is also metadata of a wheel that we would like to set. Because
of this, this setup.py file includes some related complexity.

Relevant reference documentation:
- https://setuptools.pypa.io/en/latest/userguide/extension.html#adding-commands
"""
import errno
import os
import subprocess
import sys

# distutils is deprecated but we need to keep using these imports, see
# https://github.com/pypa/setuptools/issues/2928
from distutils.command.build import build as _build
from distutils.command.clean import clean as _clean

from setuptools import Command, find_packages, setup
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
PROXY_SRC_DIR = os.path.join(ROOT_DIR, "dask-gateway-proxy")
PROXY_TGT_DIR = os.path.join(ROOT_DIR, "dask_gateway_server", "proxy")
PROXY_TGT_EXE = os.path.join(PROXY_TGT_DIR, "dask-gateway-proxy")


class build_proxy(Command):
    """
    A command the compile the golang code in ./dask-gateway-proxy and copy the
    executable binary to dask_gateway_server/proxy/dask-gateway-proxy to be
    picked up via package_data.

    setuptools.Command derived classes are required to implement
    initialize_options, finalize_options, and run.
    """

    description = "build go artifact"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.mkpath(PROXY_TGT_DIR)
        try:
            code = subprocess.call(
                ["go", "build", "-o", PROXY_TGT_EXE, "./cmd/dask-gateway-proxy"],
                cwd=PROXY_SRC_DIR,
            )
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                self.warn(
                    "Building dask-gateway-server requires a go compiler, "
                    "which wasn't found in your environment. Please install "
                    "go using a package manager (e.g. apt, brew, conda, etc...), "
                    "or download from https://go.dev/dl/."
                )
                sys.exit(1)
            raise
        if code:
            sys.exit(code)


class build_proxy_mixin:
    """
    This mixin class helps us ensure that we build the dask-gateway-proxy
    executable binary no matter if we run the: install, build, develop, or
    bdist_wheel command.
    """

    def run(self):
        if (
            not os.environ.get("DASK_GATEWAY_SERVER__NO_PROXY")
            and not getattr(self, "uninstall", False)
            and not os.path.exists(PROXY_TGT_EXE)
        ):
            self.run_command("build_proxy")
        super().run()


class build(build_proxy_mixin, _build):
    pass


class install(build_proxy_mixin, _install):
    pass


class develop(build_proxy_mixin, _develop):
    pass


class bdist_wheel(build_proxy_mixin, _bdist_wheel):
    """
    When we build wheels, they are without this override named
    "<something>-py3-none-any.whl" which is incorrect as we have a platform
    dependency.

    By overriding the `root_is_pure` option and the `get_tag` we can declare
    a dependency to the Golang compiled executable binary.

    This is based on https://stackoverflow.com/a/45150383/2220152.
    """

    def finalize_options(self):
        super().finalize_options()
        # Cleanup binaries ahead of time to avoid packaging an existing
        # binary with the wheel that shouldn't be packaged.
        if os.path.exists(PROXY_TGT_EXE):
            os.remove(PROXY_TGT_EXE)

        if os.environ.get("DASK_GATEWAY_SERVER__NO_PROXY"):
            self.root_is_pure = True
        else:
            self.root_is_pure = False

    def get_tag(self):
        python, abi, plat = super().get_tag()
        python, abi = "py3", "none"

        if os.environ.get("DASK_GATEWAY_SERVER__NO_PROXY"):
            return python, abi, "any"

        # If GOOS or GOARCH are set, we will probably be cross compiling.
        # Below we act intelligently based on a few combinations of GOOS and
        # GOARCH to provide a platform tag that is accepted by PyPI.
        #
        # PyPI restricts us to using the following platform tags:
        # https://github.com/pypa/warehouse/blob/82815b06d9f98deed5f205c66e054de59d22a10d/warehouse/forklift/legacy.py#L108-L180
        #
        # For reference, GitHub Actions available operating systems:
        # https://github.com/actions/virtual-environments#available-environments
        #
        go_plat = os.environ.get("GOOS", "") + "_" + os.environ.get("GOARCH", "")
        go_plat_to_pypi_plat_mapping = {
            "linux_amd64": "manylinux_2_17_x86_64",
            "linux_arm64": "manylinux_2_17_aarch64",
            "darwin_amd64": "macosx_10_15_x86_64",
            "darwin_arm64": "macosx_11_0_arm64",
        }
        if go_plat in go_plat_to_pypi_plat_mapping:
            plat = go_plat_to_pypi_plat_mapping[go_plat]

        return python, abi, plat


class clean(_clean):
    def run(self):
        super().run()
        if self.all and os.path.exists(PROXY_TGT_EXE):
            os.remove(PROXY_TGT_EXE)


# cmdclass describes commands we can run with "python setup.py <command>". We
# have overridden several command classes by mixing in a common capability of
# compiling the golang executable binary.
#
# It seems that only by overriding multiple command classes we can ensure we
# manage to compile the Golang executable binary before its needed. Below are
# examples of end user commands that are impacted by our changes to cmdclass.
#
#   python setup.py [bdist_wheel|build|install|develop|clean]
#   python -m pip install [--editable] .
#   python -m build --wheel .
#
cmdclass = {
    "bdist_wheel": bdist_wheel,
    "build": build,
    "build_proxy": build_proxy,
    "install": install,
    "clean": clean,
    "develop": develop,
}

if os.environ.get("DASK_GATEWAY_SERVER__NO_PROXY"):
    package_data = {}
else:
    # FIXME: Does it make sense to put the binary next to the source code
    #        (proxy/dask-gateway-proxy)? If we build a wheel now we get a folder in
    #        the wheel called purelib that includes this and it seems contradictory.
    #
    #        An action point to resolve this would be to investigate how other
    #        projects have bundled compiled binaries.
    #
    package_data = {"dask_gateway_server": ["proxy/dask-gateway-proxy"]}

# determine version from _version.py
ns = {}
with open(os.path.join(ROOT_DIR, "dask_gateway_server", "_version.py")) as f:
    exec(f.read(), {}, ns)
    VERSION = ns["__version__"]

with open("requirements.txt") as f:
    install_requires = [l for l in f.readlines() if not l.startswith("#")]

extras_require = {
    # pykerberos is tricky to install and requires a system package to
    # successfully compile some C code, on ubuntu this is libkrb5-dev.
    "kerberos": ["pykerberos"],
    "jobqueue": ["sqlalchemy>=2.0.0"],
    "local": ["sqlalchemy>=2.0.0"],
    "yarn": ["sqlalchemy>=2.0.0", "skein >= 0.7.3"],
    "kubernetes": ["kubernetes_asyncio"],
    "all_backends": [
        "sqlalchemy>=2.0.0",
        "skein >= 0.7.3",
        "kubernetes_asyncio",
    ],
}


# setup's keyword reference:
# https://setuptools.pypa.io/en/latest/references/keywords.html
#
setup(
    name="dask-gateway-server",
    version=VERSION,
    cmdclass=cmdclass,
    maintainer="Jim Crist-Harif",
    maintainer_email="jcristharif@gmail.com",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Systems Administration",
        "Programming Language :: Python :: 3",
    ],
    keywords="dask hadoop kubernetes HPC distributed cluster",
    description=(
        "A multi-tenant server for securely deploying and managing "
        "multiple Dask clusters."
    ),
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    url="https://gateway.dask.org/",
    project_urls={
        "Documentation": "https://gateway.dask.org/",
        "Source": "https://github.com/dask/dask-gateway/",
        "Issue Tracker": "https://github.com/dask/dask-gateway/issues",
    },
    packages=find_packages(),
    package_data=package_data,
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires=">=3.9",
    entry_points={
        "console_scripts": [
            "dask-gateway-server = dask_gateway_server.app:main",
            (
                "dask-gateway-jobqueue-launcher = "
                "dask_gateway_server.backends.jobqueue.launcher:main"
            ),
        ]
    },
    zip_safe=False,
)
